﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Net.WebSockets;
using System.Threading;
using System.Threading.Tasks;

namespace WingNotificationModule.Channel.WebSocket
{
    public class WebSocketIO : IDisposable
    {
        public const int BufferSize = 8192;
        private System.Net.WebSockets.WebSocket _webSocket;


        /// <summary>
        /// WebSocket connection state
        /// </summary>
        public WebSocketProcessState State
        {
            get
            {
                Enum.TryParse(_webSocket.State.ToString(), true, out WebSocketProcessState processState);
                return processState;
            }
        }

        /// <summary>
        /// IsConnection state 
        /// </summary>
        public bool IsDisconnected 
        {
            get 
            {
                return
                ( WebSocketProcessState.CloseReceived
                | WebSocketProcessState.Closed
                | WebSocketProcessState.Aborted
                | WebSocketProcessState.CloseSent).HasFlag(State);
            }
        }

        public WebSocketIO(System.Net.WebSockets.WebSocket webSocket)
        {
            _webSocket = webSocket;
        }

        /// <summary>
        /// Dispose
        /// </summary>
        public void Dispose()
        {
            _webSocket.Abort();
            _webSocket.Dispose();
        }

        /// <summary>
        /// Receive data
        /// </summary>
        /// <returns></returns>
        public async Task<WebSocketMessage> ReceiveAsync()
        {
            bool isReceiveOver = false;
            WebSocketMessageType messageType = WebSocketMessageType.Close;
            List<byte> bytes = new List<byte>();
            while (!isReceiveOver)
            {
                if (_webSocket.State != WebSocketState.Connecting && !IsDisconnected)
                {
                    var buffer = System.Net.WebSockets.WebSocket.CreateServerBuffer(BufferSize);
                    var result = await _webSocket.ReceiveAsync(buffer, CancellationToken.None);
                    isReceiveOver = result.EndOfMessage;
                    messageType = result.MessageType;
                    bytes.AddRange(buffer.Take(result.Count));
                }
            }
            var completeBuffer = new ArraySegment<byte>(bytes.ToArray());
            bytes.Clear();
            return new WebSocketMessage(completeBuffer, messageType);
        }

        /// <summary>
        /// Receive data
        /// </summary>
        /// <param name="readSize"></param>
        /// <returns></returns>
        public async Task<WebSocketMessage> ReceiveAsync(int readSize)
        {
            WebSocketMessageType messageType = WebSocketMessageType.Close;
            var buffer = System.Net.WebSockets.WebSocket.CreateServerBuffer(readSize);
            if (_webSocket.State != WebSocketState.Connecting && !IsDisconnected)
            {
                var result = await _webSocket.ReceiveAsync(buffer, CancellationToken.None);
                messageType = result.MessageType;
            }
            return new WebSocketMessage(buffer, messageType);
        }

        /// <summary>
        /// Send data async
        /// </summary>
        /// <param name="data"></param>
        /// <param name="messageType"></param>
        /// <returns></returns>
        public async Task SendAsync(byte[] data, WebSocketMessageType messageType)
        {
            var offset = 0;
            while (offset < data.Length)
            {
                if (_webSocket.State != WebSocketState.Connecting && !IsDisconnected)
                {
                    var surplus = data.Length - offset;
                    var limitSize = surplus >= BufferSize ? BufferSize : surplus;
                    var buffer = new ArraySegment<byte>(data, 0, limitSize);
                    await _webSocket.SendAsync(buffer, messageType, surplus < BufferSize, CancellationToken.None);
                    offset += limitSize;
                }
            }

        }

        /// <summary>
        /// Send data
        /// </summary>
        /// <param name="data"></param>
        /// <param name="messageType"></param>
        /// <returns></returns>
        public void Send(byte[] data, WebSocketMessageType messageType)
        {
            var offset = 0;
            while (offset < data.Length)
            {
                if (_webSocket.State != WebSocketState.Connecting && !IsDisconnected)
                {
                    var surplus = data.Length - offset;
                    var limitSize = surplus >= BufferSize ? BufferSize : surplus;
                    var buffer = new ArraySegment<byte>(data, 0, limitSize);
                    var task =_webSocket.SendAsync(buffer, messageType, surplus < BufferSize, CancellationToken.None);
                    task.Wait(10000);
                    offset += limitSize;
                }
            }

        }
    }
}
